!********************************************************************************************
!
!	EXTENDED TAUCHEN
!
!	Modified:
! 		29 September 2016
!
!	Authors:
! 		Simone Civale, Luis Díez-Catalán, Fatih Fazilet
!
!	Reference:
! 		“Discretizing a Process with Non-zero Skewness and High Kurtosis”, 2015
! 		Simone Civale, Luis Díez-Catalán, Fatih Fazilet
!
!	Description:
!		This code calibrates a discrete Markov Chain using the method described
!		in Civale et al. (2015). This method maps a calibrated first order
!		autoregressive process with normal mixture innovations (NMAR) into a
!		discrete Markov Chain. This mapping is done with the objective of
!		obtaining a Markov process that matches as closely as possible a
!		set of specified moments of the NMAR.
!		In this application we choose variance, skewness, kurtosis and autocorrelation
!		of the NMAR.
!		The application can be changed to target other moments by appropriately
!		modifying the vector of targets and the error function
!		The code is organized in three f90 files: main, tools_module and types_module
!		Additional files in the folder are makefile and stop
!		Edit the makefile and make it to generate an executable. Write "stop" instead of "ciao" in the
!		stop file to terminate the current run of the code, storing all the output
!		found so far.

PROGRAM main

    USE types_module
	! uncomment the line below to compile with intel fortran
    !     USE ifport
    USE tools_module

    IMPLICIT NONE
    REAL(dp), DIMENSION(9)                      ::  inputpar 			! parameters of the normal mixture
    INTEGER, DIMENSION(2)                       ::  cardinalities       ! number of cardinalities of the state space
    INTEGER                                     ::  iter                ! number of iterations for amoeba to find a minimum
    INTEGER, PARAMETER                          ::  t=1000000           ! number of realizations of the continuous process
    REAL(dp), DIMENSION(t)                      ::  process             ! realizations of the continuous process
    INTEGER                                     ::  n                   ! number of states
    REAL(dp), DIMENSION(9)                      ::  Targets             ! the vector of moments
    REAL(dp), DIMENSION(:,:), ALLOCATABLE       ::  transition          ! transition matrix of the approximating markov process
    REAL(dp), DIMENSION(:,:), ALLOCATABLE       ::  states				! states of the approximating markov process
    REAL(dp), DIMENSION(:,:), ALLOCATABLE       ::  RandomRestarts      ! vector to store initial random simplex for random restarts
    REAL(dp), DIMENSION(:), ALLOCATABLE         ::  yy
    REAL(dp)                                    ::  ftol,weight
    REAL(dp), DIMENSION(100)                    ::  best
	REAL(dp), DIMENSION(9)                      ::  mm1            
    INTEGER                                     ::  i,k,h,ii         
    INTEGER                                     ::  clock_start,clock_end,clock_rate
    REAL(dp)                                    ::  elapsed_time
    CHARACTER(2)                                ::  name
    REAL(dp), DIMENSION(9)                      ::  detail             
    REAL(dp)                                    ::  temp1,minpro,maxpro,start,finish
    INTEGER                                     ::  info
    CHARACTER(4)                                ::  stopcond
    CHARACTER(3)                                ::  s2
    CHARACTER(5)                                ::  s1    
	REAL(dp), DIMENSION(:,:), ALLOCATABLE       ::  Results_vector,states_vector
    
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! INPUTS 
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	
	! 1. The PARAMETRIZATION of the NMAR process to be discretized.
	! These are the parameters for a mixture of at most 3 normals.
	! Rho is the persistency of the shock. Ensure that the mean of the error term is 0.
    inputpar(1)=0.0335638760005817_dp          ! mean1
    inputpar(2)=0.0573592172915801_dp          ! variance1
    inputpar(3)=-0.302074884005235_dp          ! mean2
    inputpar(4)=1.67485882854210_dp            ! variance2
    inputpar(5)=inputpar(3)		      		   ! mean3    
    inputpar(6)=inputpar(4)         		   ! variance3
    inputpar(7)=0.9_dp   					   ! p1
    inputpar(8)=0.05_dp  					   ! p2
    inputpar(9)=0.50_dp          			   ! rho!
	
	WRITE(s1,'(F5.3)')inputpar(9)
    s2=s1(3:5)
    
    ! 2. Specify the CARDINALITY of the state space
    ! to discretize the same NMAR with several discrete Markov Chains of different
	! cardinalities, specify a vector of cardinalities
    ! for example to obtain a 3-state and 5-state discretization,
	! make sure that the size of the vector cardinalities is 2, and specify:
    ! cardinalities=(/3,5/)
    
    cardinalities=(/5,9/)
    
    ! 3. Specify the NUMBER of RANDOM RESTARTS for the global algorithm
    ! see that the algorithm restarts for a total of 2*k times
    ! k during the initial global non-focused search
    ! and k for the final focused search
    
    k=200 ! Must be a multiple of 100
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    ! here we compute the moments of the NMAR analitically
	! We do not target the mean of the process and of e as it is not necessary.
	! In fact at the end of the computation  
	! we shift the state space to ensure that the mean of the process is zero.
	! In any application where the mean of the process is not zero, the
	! state space can be shifted to ensure matching the mean of the process
	! without affecting the other centered moments targeted by this calibration.
    Targets=ParToMom(inputpar)
	! In this application the moments are
    ! 1 mean of the process
    ! 2 variance of the process
    ! 3 skewness of the process
    ! 4 kurtosis of the process
    ! 5 mean of the e
    ! 6 variance of the e
    ! 7 skewness of the e
    ! 8 kurtosis of the e
    ! 9 rho
	
	
    ! The process is simulated and stored
	! The purpose of the simulation is not to compute the moments of the NMAR
	! But we will use the percentiles of the process to establish what's the
	! support fir the initial guesses of the state space.
    CALL draw_process(inputpar,process)
    ! let's sort the realization of the process
    CALL dlasrt('I',t,process,info)

	! let's print the moments
    CALL PrintMoments()
   
   	! this is the main loop, each loop performes a global search for
	! the optimal state space for a given cardinality of the state space
    DO ii=1,size(cardinalities) 
		CALL InitiateLoop()
        ! the following loop runs the first half of the global optimization
		! k restarts are considered and from each restarting point
		! the amoeba algorithm finds a local minimum for the discrepancy function
        DO h=1,k
			! To stop the program run and store the results produced so far
			! edit the stop file in the folder and write "stop" instead of "ciao"
			! This can be useful as you might set k to 100M restarts but you might
			! realize that after fewer iteration the discrete process already attains
			! exactly the moments you want to match
            OPEN(UNIT=2,FILE='stop')
            READ(2,'(A4)') stopcond
            CLOSE(2)
            IF (stopcond .EQ. 'stop') EXIT
			! produce the  random simplex to restart the optimization
            RandomRestarts=RandomSimplex(n,minpro,maxpro)
            ! set the initial simplex for the optimization algorithm
            states=RandomRestarts
			! compute for each guess of the state space the 
			! associated discrepancy
            DO i=1,n+1 
                yy(i)=ErrorFunction(inputpar,targets,states(i,:)) 
            END DO              
            ! optimize to find the best state space
            CALL amoeba(states,yy,ftol,iter,inputpar,targets)						
            ! store the results of the optimization in the best vector
            IF (best(n+1) .GT. yy(1)) THEN
                best(1:n)=states(1,:)
                best(n+1)=yy(1)									
            END IF			
			! every 100 restarts print the results of the global
			! optimization
			IF (MOD(h,100) .EQ. 0) THEN
				finish=0.0_dp
				CALL CPU_TIME(finish)
				temp1=PerError(inputpar,targets,best(1:n),detail)
				Results_vector(h/100,1)=REAL(h)
				Results_vector(h/100,2)=best(n+1)
				Results_vector(h/100,3)=temp1
				Results_vector(h/100,4)=finish-start
				states_vector(h/100,:)=best(1:n)	
				PRINT'(F12.1,2F15.8,F15.1)', Results_vector(h/100,:)	
			END IF							    
        END DO
        ! the following loop runs the second half of the global optimization
		! k restarts are considered and from each restarting point
		! the amoeba algorithm finds a local minimum for the discrepancy function		
		! in this second half of the global algorithm the restarts zoom in
		! towards the best state space found so far
        DO h=1,k
            OPEN(UNIT=2,FILE='stop')
            READ(2,'(A4)') stopcond
            CLOSE(2)
            IF (stopcond .EQ. 'stop') EXIT  
            RandomRestarts=RandomSimplex(n,minpro,maxpro)   
			! in this part of the loop the weight specifies how the 
			! random simplex and the best so far are linearly combined together
			! as h goes to k the weight goes to 1, meaning that the restarts
			! zoom in the best state space found so far     
            weight=(real(k-h+1)/real(k))
            DO i=1,n+1
                states(i,:)=RandomRestarts(i,:)*weight+best(1:n)*(1.0-weight)
            END DO
            DO i=1,n+1 
                yy(i)=ErrorFunction(inputpar,targets,states(i,:))
            END DO    
            CALL amoeba(states,yy,ftol,iter,inputpar,targets)           
			IF (best(n+1) .GT. yy(1)) THEN
                best(1:n)=states(1,:)
				best(n+1)=yy(1)
            END IF
			IF (MOD(h,100) .EQ. 0) THEN
				finish=0.0_dp
				CALL CPU_TIME(finish)
				temp1=PerError(inputpar,targets,best(1:n),detail)
				Results_vector((k+h)/100,1)=REAL(h)
				Results_vector((k+h)/100,2)=best(n+1)
				Results_vector((k+h)/100,3)=temp1
				Results_vector((k+h)/100,4)=finish-start
				states_vector((k+h)/100,:)=best(1:n)
				PRINT'(F12.1,2F15.8,F15.1)', Results_vector((k+h)/100,:)						
			END IF					
        END DO
		! loop is finalized, all the output written to files
		CALL FinalizeLoop()
		!
    END DO

	CONTAINS
		
		SUBROUTINE InitiateLoop()
			start=0.0_dp
			CALL CPU_TIME(start)
	        n=cardinalities(ii)
	        CALL SYSTEM_CLOCK(COUNT_RATE=clock_rate)
	        CALL SYSTEM_CLOCK(COUNT=clock_start)
	        ! now we set the number of states n
	        print*,'number of states=',n
			PRINT '(A12,2A15,A15)','No of iter','criteria','percent dev','time'
			! Here we set the minimum and maximum of the support for guesses
			! of the random simplex
			! To be more clear: for the global optimization we need to restart
			! the optimization at different initial locations
			! These initial point need to be chosen on some support.
			! We found a rule of thumb that is coded here about what is a good
			! choice for this support.
			! As you can see this rule of thumb involves using the percentiles
			! of the realized process.
	        IF (n .LE. 15) THEN
	            minpro=process(10*t/100)
	            maxpro=process(90*t/100)
	        ELSE
	            minpro=process(1*t/100)
	            maxpro=process(99*t/100)
	        END IF
	        ! we allocate the necessary vectors
			! simplex for ameoba
	        ALLOCATE(RandomRestarts(n+1,n)) 
			! states is equal to RandomRestarts or a convex combination of it with the best up to that point
	        ALLOCATE(states(n+1,n))			
			! is the vector of discrepancies for each node of the simplex
	        ALLOCATE(yy(n+1))				
	        ALLOCATE(transition(n,n))
			! we keep and print the results for every 100th iteration
			ALLOCATE(Results_vector(2*k/100,4)) 
			ALLOCATE(states_vector(2*k/100,n))
	        best(n+1)=1000.0_dp
	        ! we set the tolerance for the amoeba
	        ftol=10.0_dp**(-6.0_dp)			
		END SUBROUTINE InitiateLoop
		
		SUBROUTINE PrintMoments()
			PRINT *, ''
		    PRINT'(A31,F6.2,F6.2)','           mean of the process ',Targets(1)
		    PRINT'(A31,F6.2,F6.2)','       variance of the process ',Targets(2)
		    PRINT'(A31,F6.2,F6.2)','       skewness of the process ',Targets(3)
		    PRINT'(A31,F6.2,F6.2)','       kurtosis of the process ',Targets(4)
		    PRINT'(A31,F6.2,F6.2)','                 mean of the e ',Targets(5)
		    PRINT'(A31,F6.2,F6.2)','             variance of the e ',Targets(6)
		    PRINT'(A31,F6.2,F6.2)','             skewness of the e ',Targets(7)
		    PRINT'(A31,F6.2,F6.2)','             kurtosis of the e ',Targets(8)
		    PRINT'(A31,F6.2,F6.2)','                           rho ',Targets(9)
			PRINT *, ''
		END SUBROUTINE PrintMoments
		
		SUBROUTINE WriteResults()
	        WRITE (name,'(I2)') n
	        OPEN(unit=7,file='criteria_iter'//name//'_'//s2)
			OPEN(unit=9,file='states_iter'//name//'_'//s2)
			WRITE(7,'(A12,2A15,A15)') 'No of iter','criteria','percent dev','time'
			DO i=1,2*k/100
				WRITE(7,'(F12.1,2F15.8,F15.1)') Results_vector(i,:)		
				WRITE(9,'('//name//'F11.5)') states_vector(i,1:n)
			END DO
	        OPEN(unit=1,file='results'//name//'_'//s2)
	        WRITE(1,*) "number of states"
	        WRITE(1,'(I5)') n
	        WRITE(1,*) ""       
	        WRITE(1,*) "time elapsed: "
	        WRITE(1,'(F9.3)') elapsed_time
	        WRITE(1,*) ""       
	        WRITE(1,*) "criterion: "
	        WRITE(1,'(F9.3)') best(n+1)
	        WRITE(1,*) ""       
	        temp1=PerError(inputpar,targets,best(1:n),detail)
	        WRITE(1,*) "percentage criterion"
	        WRITE(1,'(F9.3)') temp1
	        WRITE(1,*) ""
			WRITE(1,*) "Everything is targeted except E(y) and E(e)"
			WRITE(1,*) ""        
	        WRITE(1,'(A31,A12,A10)') " detailed percentage criterion  ","perc. dev","target"
	        WRITE(1,'(A31,F12.4,F10.3)') ' mean of the process            ',detail(1),Targets(1)
	        WRITE(1,'(A31,F12.4,F10.3)') ' variance of the process        ',detail(2),Targets(2)
	        WRITE(1,'(A31,F12.4,F10.3)') ' skewness of the process        ',detail(3),Targets(3)
	        WRITE(1,'(A31,F12.4,F10.3)') ' kurtosis of the process        ',detail(4),Targets(4)
	        WRITE(1,'(A31,F12.4,F10.3)') ' mean of the error              ',detail(5),Targets(5)
	        WRITE(1,'(A31,F12.4,F10.3)') ' variance of the error          ',detail(6),Targets(6)
	        WRITE(1,'(A31,F12.4,F10.3)') ' skewness of the error          ',detail(7),Targets(7)
	        WRITE(1,'(A31,F12.4,F10.3)') ' kurtosis of the error          ',detail(8),Targets(8)
	        WRITE(1,'(A31,F12.4,F10.3)') ' autocorrelation of the process ',detail(9),Targets(9)
	        CALL TransitionOneStep(inputpar,best(1:n),transition)
			WRITE(1,*) ""  
	        WRITE(1,*) "transition prob"
	        WRITE(1,'('//name//'F9.5)') TRANSPOSE(transition)
			mm1=TransitionStates2Targets(transition,best(1:n))
			best(1:n)=best(1:n)-mm1(1)
	        WRITE(1,*) ""       
	        WRITE(1,*) "states"
	        WRITE(1,'(F9.5)') best(1:n)
			WRITE(1,*) ""  
			WRITE(1,*) "stationary prob"
	        WRITE(1,'('//name//'F9.5)') stationary(transition)
	        CLOSE(1)
			CLOSE(7)
			CLOSE(9)			
		END SUBROUTINE WriteResults
		
		SUBROUTINE FinalizeLoop()
	        PRINT*,""
	        CALL SYSTEM_CLOCK(COUNT=clock_end)
	        elapsed_time=REAL(clock_end-clock_start)/REAL(clock_rate)

	        ! we write the result in a file
			CALL WriteResults()

	        ! deallocating
	        DEALLOCATE(RandomRestarts)
	        DEALLOCATE(transition)
	        DEALLOCATE(states)
	        DEALLOCATE(yy)
			DEALLOCATE(Results_vector)
			DEALLOCATE(states_vector)			
		END SUBROUTINE FinalizeLoop
	
END PROGRAM

